/* 

  RMC - Ron's MUD Client
  Copyright (C) 2000 Ron Martin <rmartin@cps.cmich.edu>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

/*
	The core of this program is here.
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "client.h"
#include "interface.h"

struct conn_info * info_database = NULL;

GdkColor colors[16];
GdkColor color_normal;

/*
	main() - gets the ball rolling.
	This function reads in the database, creates
	the main window, allocates all the nice colors,
	starts gtk handlers, and continues.
*/
int main(int argc, char *argv[])
{
  GtkWidget * main_window;

  int read_info_database();
  void allocate_colors();
  
  gtk_set_locale (); 
  gtk_init (&argc, &argv);
  
  if(!read_info_database())
  {
  	gtk_main_quit();
  	exit(0);
  }
  
  main_window = create_main_window();

  allocate_colors();
  
  gtk_widget_show(main_window);
  gtk_main ();
  return 0;
}

/*
	read_info_database() - reads the database from file
	This should probably be in helpers.c.
	This just reads the database from file.  The database file
	will always be ~/.rmcrc.  The HOME directory is ascertained
	by the getenv() function (see the man page).
*/
int read_info_database()
{
	char temp[1024];
	struct conn_info * temp_database = NULL;
	struct parse_type * temp_alias_data, * temp_trigger_data, * temp_parse_data = NULL;
	struct log_info * temp_log_data = NULL;
	int name_set = FALSE;
	char * fname, * home;
	
	void remove_newline(char * input_string);
	
	FILE * rc_file;

	home = getenv("HOME"); /* get home directory */
	
	fname = g_malloc(strlen(home) + 8);
	strcpy(fname, home);
	strcat(fname, "/.rmcrc");
		
	rc_file = fopen(fname, "r");
	if(!rc_file)
	{
		rc_file = fopen(fname, "w");
		if(!rc_file) /* can't open */
		{
			error_handle(2);
			return FALSE;
		}
		return TRUE;
	}
	else
	{
		while(fgets(temp, 1024, rc_file)) /* go until EOF */
		{
		    if(temp[0] == '\n')
		    	continue;
		    	
			if(temp[0] == '{')
			{
				if(!info_database) /* no database yet? */
				{ /* then create it */
					info_database = g_malloc0(sizeof(conn_info));
					temp_database = info_database;
					temp_database->font_name = g_strdup("-misc-fixed-medium-r-semicondensed-*-*-120-*-*-c-*-iso8859-8");
					temp_database->font = gdk_font_load("-misc-fixed-medium-r-semicondensed-*-*-120-*-*-c-*-iso8859-8");
				}
				else /* yes, there is a database */
				{
					temp_database->next = g_malloc0(sizeof(conn_info));
					temp_database = temp_database->next;
					temp_database->font_name = g_strdup("-misc-fixed-medium-r-semicondensed-*-*-120-*-*-c-*-iso8859-8");
					temp_database->font = gdk_font_load("-misc-fixed-medium-r-semicondensed-*-*-120-*-*-c-*-iso8859-8");
				}
				
				if(!(fgets(temp, 1024, rc_file)))
				{
					error_handle(1); /* the file is messed up. */
					return FALSE;
				}
				
				temp_database->mud_name = g_malloc(strlen(temp) + 1);
				strcpy(temp_database->mud_name, temp);
				remove_newline(temp_database->mud_name);
				
				if(!(fgets(temp, 1024, rc_file)))
				{
					error_handle(1); /* the file is messed up. */
					return FALSE;
				}
				
				temp_database->host_name = g_malloc(strlen(temp) + 1);
				strcpy(temp_database->host_name, temp);
				remove_newline(temp_database->host_name);
								
				if(!(fgets(temp, 1024, rc_file)))
				{
					error_handle(1); /* the file is messed up. */
					return FALSE;
				}
				
				temp_database->port = atoi(temp);
				temp_alias_data = NULL;
				temp_trigger_data = NULL;
				
				while(fgets(temp, 1024, rc_file))
				{
					if(temp[0] == '}')
						break;
						
					if(temp[0] == '\n')
						continue;
						
					if(temp[0] == '-')
					{
						temp_database->font_name = g_strdup(temp);
						remove_newline(temp_database->font_name);
						temp_database->font = gdk_font_load(temp_database->font_name);
						continue;
					}
						
					if(!strcmp(temp, "log\n"))
					{
						if(!temp_database->logs)
						{
							temp_database->logs = g_malloc0(sizeof(log_info));
							temp_log_data = temp_database->logs;
						}
						else
						{
							temp_log_data->next = g_malloc0(sizeof(log_info));
							temp_log_data = temp_log_data->next;
						}
						if(!fgets(temp, 1024, rc_file))
						{
							error_handle(1);
							return FALSE;
						}
						temp_log_data->name = g_strdup(temp);
						remove_newline(temp_log_data->name);
						
						if(!fgets(temp, 1024, rc_file))
						{
							error_handle(1);
							return FALSE;
						}
						if(temp[0] != '*')
						{
							temp_log_data->log_file = g_strdup(temp);
							remove_newline(temp_log_data->log_file);
						}
						else
							temp_log_data->log_file = NULL;
						
						if(!fgets(temp, 1024, rc_file))
						{
							error_handle(1);
							return FALSE;
						}
						if(!strcmp(temp, "True\n"))
							temp_log_data->log_to_window = TRUE;
						else
							temp_log_data->log_to_window = FALSE;
						if(!fgets(temp, 1024, rc_file))
						{
							error_handle(1);
							return FALSE;
						}
						if(strcmp(temp, "done\n"))
						{
							error_handle(1);
							return FALSE;
						}
						continue;
					}

					if(!strcmp(temp, "alias\n"))
					{
						if(!temp_database->aliases) /* are there aliases? */
						{
							temp_database->aliases = g_malloc0(sizeof(parse_type));
							temp_alias_data = temp_database->aliases;
						}
						else /* yes, there are */
						{
							temp_alias_data->next = g_malloc0(sizeof(parse_type));
							temp_alias_data = temp_parse_data->next;
							temp_alias_data->next = NULL;
						}
						temp_parse_data = temp_alias_data;
					}
					else if(!strcmp(temp, "trigger\n"))
					{
						if(!temp_database->triggers) /* are there triggers? */
						{
							temp_database->triggers = g_malloc0(sizeof(parse_type));
							temp_trigger_data = temp_database->triggers;
						}
						else /* yes there are */
						{
							temp_trigger_data->next = g_malloc0(sizeof(parse_type));
							temp_trigger_data = temp_parse_data->next;
							temp_trigger_data->next = NULL;
						}
						temp_parse_data = temp_trigger_data;
					}
					else
					{
						error_handle(3); /* the file is messed up. */
						return FALSE;
					}
					
					if(!(fgets(temp, 1024, rc_file)))
					{
						error_handle(1); /* the file is messed up. */
						return FALSE;
					}
					
					temp_parse_data->input_string = g_malloc(strlen(temp) + 1);
					strcpy(temp_parse_data->input_string, temp);
					
					while(1)
					{
						if(!(fgets(temp, 1024, rc_file)))
						{
							error_handle(1); /* the file is messed up. */
							return FALSE;
						}
						if(!strcmp(temp, "done\n"))
							break;
						if(!name_set)
						{
							temp_parse_data->output_string = g_malloc(strlen(temp) + 1);
							strcpy(temp_parse_data->output_string, temp);
							name_set = TRUE;
						}
						else
						{
							temp_parse_data->output_string = g_realloc(temp_parse_data->output_string, strlen(temp) + strlen(temp_parse_data->output_string) + 1);
							/* ^^ this is the longest line I ever typed */
							strcat(temp_parse_data->output_string, temp);
						}
					}
					name_set = FALSE;
				} /* lots of braces.  Confusing */
			}
		}
    	}
	fclose(rc_file); /* all done */
	g_free(fname);
	return TRUE;
}

/*
	error_handle() - given an error number, display a message
	This handles all errors.  If you need to send an error to
	the user, this is the function for you.  This should probably
	be in helper.c, but oh well.  A lot of these error should never
	happen, but it doesn't hurt to check for them anyway.
*/
int error_handle(int err_number)
{
	GtkWidget * error_window;
	char text[256], text2[1024];
	int show_window = TRUE;
		
	switch(err_number)
	{
		case 1: /* database file is foobar */
			strcpy(text, "FATAL - Corrupt resource file");
			strcpy(text2, "The resource file, ~/.rmcrc, is corrupt.\nIt is recommended that this file is deleted.\nIt is also recommended that, although possible,the resource file should\nnot be edited by hand.\n\nThe application will now close.\n");
			show_window = FALSE;
			break;
		case 2: /* no permission to open file */
			strcpy(text, "FATAL - Cannot open resource file");
			strcpy(text2, "The resource file,~/.rmcrc, cannot be opened by the application.\nBe sure you have the proper permissions to read from this file.\nThis error will not be generated if the resource file is missing, but, instead,\na new file will be created.\n\nThe application will now close.\n");
			show_window = FALSE;
			break;
		case 3: /* resource file is messed up a tad. */
			strcpy(text, "FATAL - Unknown command in resource file");
			strcpy(text2, "The application has attempted to parse a command in the resource file,\nbut was unable to do so. It is recommended that the resource file be deleted.\nIt is also recommended that, although possible, the resource file should not be edited by hand.\n\nThe application will now close.\n");
			show_window = FALSE;
			break;
		case 4: /* no write permission for file */
			strcpy(text, "ERROR - Cannot write to resource file");
			strcpy(text2, "The resource file,~/.rmcrc, cannot be opened for writing.\nBe sure you have the proper permissions to access this file.\n\nThe application will continue, but any changes made will not be\nsaved to disk for future sessions.\n");
			show_window = FALSE;
			break;
		case 5: /* host is gone or not a fully-qualified domain name */
			strcpy(text, "ERROR - Cannot connect to host");
			strcpy(text2, "\nThe application is unable to connect to the specified host.\nThe host server could be down, but make sure the\nhost name and the port number are correct.\n");
			break;
		case 6: /* disconnect */
		    strcpy(text, "ALERT - Disconnected from host");
		    strcpy(text2, "\nThe connection to the host has been lost.\nThis can be due to a number of reasons:\n\n -You have quit\n -You have lost your network connection\n -The mud crashed\n -You have been disconnected by an immortal.\n");
		    break;
		case 7: /* no host name or port (or both!) */
			strcpy(text, "ALERT - Need host name and port number");
			strcpy(text2, "\nAll connection database entries require\na host name and a port number.\n");
			break;
		case 8: /* duplicate connection */
			strcpy(text, "ALERT - Connection already defined");
			strcpy(text2, "\nThe connection information you have supplied is\nalready in the connection database.\nIf you wish to change this info,\nplease use the \"edit\" \nfeature of the application.\n");
			break;
		case 9: /* color allocation failed */
			strcpy(text, "ALERT - Unable to allocate color");
			strcpy(text2, "\nThe application was unable to get a needed color\nfrom the color map.\nThough this is not fatal, it could cause some\ntext to have strange colors.\n");
		case 10: /* Uh-oh.  Bad error. */
			strcpy(text, "ERROR - Deletion of imaginary data");
			strcpy(text2, "\nOops.  The application tried to delete data that does not exist.\nEither there is a bug, or the computer is smoking crack.\nIn either case, contact me:\n\n rmartin@cps.cmich.edu\n");
			break;
		case 11: /* bad linked list */
			strcpy(text, "ERROR - Connection/database mismatch");
			strcpy(text2, "\nThe application attempted to delete some information,\nbut could not find its predecessor in the linked list,\nand was not the first element in the list.\nI can tell you really don't care, so email me:\n\n rmartin@cps.cmich.edu\n");
			break;
		case 12: /* already connected */
			strcpy(text, "ALERT - Already connected");
			strcpy(text2, "\nThe application already has a conenction\nestablished with the selected information.\nEither:\n\n 1) Disconnect\n 2) Create a new connection\n");
			break;
		case 13: /* already editing */
			strcpy(text, "ALERT - Already editing");
			strcpy(text2, "\nYou are already editing this connection.\n");
			break;
		case 14: /* already opened */
			strcpy(text, "ALERT - Already opened");
			strcpy(text2, "\nThis connection is already opened.\n");
			break;
		case 15: /* still conencted. */
			strcpy(text, "ALERT - Still connected");
			strcpy(text2, "\nThis connection is still live.\nPlease disconnect first.\n");
			break;
		case 16: /* quitting while connection open. */
			strcpy(text, "ALERT - Cannot quit");
			strcpy(text2, "\nA connection is still alive.\nPlease disconnect all\nopen connections and try\nto quit again.\n");
			break;
		case 17: /* Invalid meta (#) command */
			strcpy(text, "ALERT - Invalid meta-command");
			strcpy(text2, "\nThe command you typed begins with\na pound sign (#), but is not\na legal meta-command.  Legal\nmeta-commands are:\n #reconnect\n #disconnect\n #close\n #edit\n #log\n #help\n");
			break;
		case 18: /* can't open log file */
			strcpy(text, "ERROR - Cannot open log file");
			strcpy(text2, "\nThe application was unable\nto open a specified log file.\nPlease be sure you have proper\npermissions to write to all your log\nfiles.\nThis error is not fatal, but furthur\nattmempts will produce this error message.\n");
			break;
		case 19: /* Log widget puked.  Yuk. */
			strcpy(text, "ERROR - Missing widget");
			strcpy(text2, "\nThe application attempted to write\nsome text to an imaginary window.\nPlease contact the author at:\nrmartin@cps.cmich.edu\n");
			break;
		case 20: /* Can't open mud list for reading. */
			strcpy(text, "Error - Cannot open mud list");
			strcpy(text2, "\nThe application attempted to open\nthe MUD list, but was not able to.\nBe sure the list is in the correct path and\nyou have proper permissions to read\nfrom the list.\n");
			break;
		case 21: /* Can't establish connection to server */
			strcpy(text, "Error - Cannot connect to list server");
			strcpy(text2, "\nThe application cannot connect\nto the MUD list server\n\nwww.mudconnector.com\n\nThe server may be busy or down.\nTry again later.\n");
			break;
		case 22: /* Can't open mud list for write or append. */
			strcpy(text, "Error - Cannot create MUD list");
			strcpy(text2, "\nThe application cannot create or\nappend to the local file that\nholds the MUD list (~/.rmclist).\nPlease make sure you have the\nproper permissions to write to this file.\n");
		case 23: /* The header info is incorrect.  Thanks for playing. */
			strcpy(text, "Error - Invalid MUD list header info");
			strcpy(text2, "\nWhile attempting to parse the header\ninfo for the MUD list, the\napplication got confused and decided to\nforget the whole download thing.\nPlease e-mail me (rmartin@cps.cmich.edu)\nand let me know when you tried\nto update the MUD list and anything\nelse you think might be important.\n\nSorry about this.\n");
		default: /* how did we get here?? */
			strcpy(text, "FATAL - Oops");
			strcpy(text2, "\nI don't know what happened,\nbut you're screwed,\nwhatever it was.\n\nRon Martin: rmartin@cps.cmich.edu\n");
			show_window = FALSE;
	}
	if(show_window) /* if gtk_init() hasn't been called, then we can't make a window. */
	{
		error_window = create_error_window(text, text2);
	
		gtk_widget_show(error_window);
	}
	else /* this will not exist if run from an icon or something. */
		g_print("*[%s]*\n\n%s", text, text2);
		
	return TRUE;
}

/*
	new_socket() - initialize and connect a new socket.
	This uses the UNIX socket functions to create a new
	socket.  This should probably be in a whole new file
	called net.c, but I'm not worried about it.
*/
int new_socket(char * host_name, int port)
{
	struct hostent * server_host_name;
	struct sockaddr_in sin;
	int socket_descriptor;
	
	void remove_newline(char * input_string);
	
	remove_newline(host_name);
	
	if(!(server_host_name = gethostbyname(host_name)))
	{
		error_handle(5); /* no host or not FQDN */
		return 0;
	}
	
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_addr.s_addr = ((struct in_addr *)(server_host_name->h_addr))->s_addr;
	sin.sin_port = htons(port);
	
	socket_descriptor = socket(AF_INET, SOCK_STREAM, 0); /* make the socket */
	
	if(connect(socket_descriptor, (void *)&sin, sizeof(sin)))
	{
		error_handle(5); /* can't connect */
		return 0;
	}
	
	return socket_descriptor;
}

/*
	allocate_colors() - Creates the color table and
	initializes the colors{} aggregation.
	This should probably be in helpers.c.
*/
void allocate_colors()
{
	int i;
	GdkColormap * cmap;
	
	cmap = gdk_colormap_get_system();
	
	color_normal.red = 0xDDDD;
	color_normal.green = 0xDDDD;
	color_normal.blue = 0xDDDD;

	for(i = 0; i < 16; i++)
	{
		switch(i)
		{
			case 0: /* black */
				colors[i].red = 0;
				colors[i].green = 0;
				colors[i].blue = 0;
				break;
			case 1: /* red */
				colors[i].red = 0xDDDD;
				colors[i].green = 0;
				colors[i].blue = 0;
				break;
			case 2: /* green */
				colors[i].red = 0;
				colors[i].green = 0xDDDD;
				colors[i].blue = 0;
				break;
			case 3: /* orangeish */
				colors[i].red = 0xDDDD;
				colors[i].green = 0xDDDD;
				colors[i].blue = 0;
				break;
			case 4: /* blue */
				colors[i].red = 0;
				colors[i].green = 0;
				colors[i].blue = 0xDDDD;
				break;
			case 5: /* purpleish */
				colors[i].red = 0xDDDD;
				colors[i].green = 0;
				colors[i].blue = 0xDDDD;
				break;
			case 6: /* blue-greenish */
				colors[i].red = 0;
				colors[i].green = 0xDDDD;
				colors[i].blue = 0xDDDD;
				break;
			case 7: /* white */
				colors[i].red = 0xBBBB;
				colors[i].green = 0xBBBB;
				colors[i].blue = 0xBBBB;
				break;
			case 8: /* grey */
				colors[i].red = 0xAAAA;
				colors[i].green = 0xAAAA;
				colors[i].blue = 0xAAAA;
				break;
			case 9: /* bright red */
				colors[i].red = 0xffff;
				colors[i].green = 0;
				colors[i].blue = 0;
				break;
			case 10: /* bright green */
				colors[i].red = 0;
				colors[i].green = 0xffff;
				colors[i].blue = 0;
				break;
			case 11: /* bright orangeish */
				colors[i].red = 0xffff;
				colors[i].green = 0xffff;
				colors[i].blue = 0;
				break;
			case 12: /* bright blue */
				colors[i].red = 0;
				colors[i].green = 0;
				colors[i].blue = 0xffff;
				break;
			case 13: /* bright purpleish */
				colors[i].red = 0xffff;
				colors[i].green = 0;
				colors[i].blue = 0xffff;
				break;
			case 14: /* bright blue-greenish */
				colors[i].red = 0;
				colors[i].green = 0xffff;
				colors[i].blue = 0xffff;
				break;
			case 15: /* bright white */
				colors[i].red = 0xffff;
				colors[i].green = 0xffff;
				colors[i].blue = 0xffff;
				break;
		}
		
		if (!gdk_color_alloc(cmap, &colors[i]))
		{
			error_handle(9); /* coudn't allocate */
			return;
		}
	}
	
	if(!gdk_color_alloc(cmap, &color_normal))
	{
		error_handle(9); /* couldn't allocate */
		return;
	}
}

/*
	remove_newline() - removes newlines from a string
	I didn't even know I wrote this until I was going over
	my comments.  It just changes the newlines to NULL.
*/
void remove_newline(char * input_string)
{
	int i = 0;
	while(input_string[i] != '\0')
	{
		if(input_string[i] == '\n')
		{
			input_string[i] = '\0';
			break;
		}
		i++;
	}
}
